"""Implements the neck of YOLOv4, including the SPP and the modified PAN"""
import tensorflow as tf

from .layers import conv_bn


def yolov4_neck(input_shapes):
    """
    Implements the neck of YOLOv4, including the SPP and the modified PAN.

    Args:
        input_shapes (List[Tuple[int]]): List of 3 tuples, which are the output shapes of the backbone.
            For CSPDarknet53, those are: [(52, 52, 256), (26, 26, 512), (13, 13, 1024)] for a (416, 416) input.

    Returns:
        tf.keras.Model: Neck model
    """
    input_1 = tf.keras.Input(shape=filter(None, input_shapes[0]))
    input_2 = tf.keras.Input(shape=filter(None, input_shapes[1]))
    input_3 = tf.keras.Input(shape=filter(None, input_shapes[2]))

    x = conv_bn(input_3,
                filters=512,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=1024,
                kernel_size=3,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=512,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")

    maxpool_1 = tf.keras.layers.MaxPool2D((5, 5), strides=1, padding="same")(x)
    maxpool_2 = tf.keras.layers.MaxPool2D((9, 9), strides=1, padding="same")(x)
    maxpool_3 = tf.keras.layers.MaxPool2D((13, 13), strides=1,
                                          padding="same")(x)

    spp = tf.keras.layers.Concatenate()([maxpool_3, maxpool_2, maxpool_1, x])

    x = conv_bn(spp,
                filters=512,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=1024,
                kernel_size=3,
                strides=1,
                activation="leaky_relu")
    output_3 = conv_bn(x,
                       filters=512,
                       kernel_size=1,
                       strides=1,
                       activation="leaky_relu")
    x = conv_bn(output_3,
                filters=256,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")

    upsampled = tf.keras.layers.UpSampling2D()(x)

    x = conv_bn(input_2,
                filters=256,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")
    x = tf.keras.layers.Concatenate()([x, upsampled])

    x = conv_bn(x,
                filters=256,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=512,
                kernel_size=3,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=256,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=512,
                kernel_size=3,
                strides=1,
                activation="leaky_relu")
    output_2 = conv_bn(x,
                       filters=256,
                       kernel_size=1,
                       strides=1,
                       activation="leaky_relu")
    x = conv_bn(output_2,
                filters=128,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")

    upsampled = tf.keras.layers.UpSampling2D()(x)

    x = conv_bn(input_1,
                filters=128,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")
    x = tf.keras.layers.Concatenate()([x, upsampled])

    x = conv_bn(x,
                filters=128,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=256,
                kernel_size=3,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=128,
                kernel_size=1,
                strides=1,
                activation="leaky_relu")
    x = conv_bn(x,
                filters=256,
                kernel_size=3,
                strides=1,
                activation="leaky_relu")
    output_1 = conv_bn(x,
                       filters=128,
                       kernel_size=1,
                       strides=1,
                       activation="leaky_relu")

    return tf.keras.Model([input_1, input_2, input_3],
                          [output_1, output_2, output_3],
                          name="YOLOv4_neck")



"""
Download util from Google Drive.

Taken from: https://stackoverflow.com/questions/38511444/python-download-files-from-google-drive-using-url
"""
import requests

import tensorflow as tf


def download_file_from_google_drive(file_id, destination, target_size=None):
    """Download file by id from Google Drive"""
    google_drive_base_url = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(google_drive_base_url,
                           params={"id": file_id},
                           stream=True)
    token = get_confirm_token(response)

    if token:
        params = {"id": file_id, "confirm": token}
        response = session.get(google_drive_base_url,
                               params=params,
                               stream=True)

    save_response_content(response, destination, target_size)


def get_confirm_token(response):
    """Get validation token from Google Drive request"""
    for key, value in response.cookies.items():
        if key.startswith("download_warning"):
            return value

    return None


def save_response_content(response, destination, target_size=None):
    """Save response content to file"""
    chunk_size = 32768

    progbar = tf.keras.utils.Progbar(
        target_size) if target_size is not None else None

    with open(destination, "wb") as file:
        for chunk_index, chunk in enumerate(response.iter_content(chunk_size)):
            if chunk:  # filter out keep-alive new chunks
                file.write(chunk)
                if progbar:
                    progbar.update(chunk_index * chunk_size)



"""
Contains original YOLO anchors and associated utils
"""
import numpy as np

YOLOV4_ANCHORS = [
    np.array([(12, 16), (19, 36), (40, 28)], np.float32),
    np.array([(36, 75), (76, 55), (72, 146)], np.float32),
    np.array([(142, 110), (192, 243), (459, 401)], np.float32),
]

YOLOV3_ANCHORS = [
    np.array([(10, 13), (16, 30), (33, 23)], np.float32),
    np.array([(30, 61), (62, 45), (59, 119)], np.float32),
    np.array([(116, 90), (156, 198), (373, 326)], np.float32),
]


def compute_normalized_anchors(anchors, input_shape):
    """
    Compute anchors resizing based on the architecture input shapes
    Args:
        anchors (List[numpy.array[int, 2]]): List of 3 numpy arrays containing the anchor sizes used for
            each stage. The first and second columns of the numpy arrays respectively contain the anchors width and
            height.
        input_shape (Tuple[int]): Input shape of the Network

    Returns:
        (List[numpy.array[int, 2]]): anchors resized based on the input shape of the Network.
    """
    height, width = input_shape[:2]
    return [anchor / np.array([width, height]) for anchor in anchors]



"""Common layer architecture such as Conv->BN->Mish or Conv->BN->LeakyReLU"""
import tensorflow as tf
import tensorflow_addons as tfa


def conv_bn(
    inputs,
    filters,
    kernel_size,
    strides,
    padding="same",
    zero_pad=False,
    activation="leaky",
):
    """
    Applies successively Conv2D -> BN -> LeakyReLU

    Args:
        inputs (tf.Tensor): 4D (N,H,W,C) input tensor
        filters (int): Number of convolutional filters
        kernel_size (int): Size of the convolutional kernel
        strides (int): Strides used for the convolution
        padding (str): Type of padding used in the convolution
        zero_pad (bool): If true, will zero-pad the input
        activation (string): Activation layer. Can be "mish" or "leaky_relu", or linear otherwise

    Returns:
        tf.Tensor: 4D (N,H,W,C) output tensor
    """
    if zero_pad:
        inputs = tf.keras.layers.ZeroPadding2D(((1, 0), (1, 0)))(inputs)

    x = tf.keras.layers.Conv2D(
        filters=filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        use_bias=False,
    )(inputs)
    x = tf.keras.layers.BatchNormalization()(x)
    if activation == "leaky_relu":
        x = tf.keras.layers.LeakyReLU(alpha=0.1)(x)
    elif activation == "mish":
        x = tfa.activations.mish(x)

    return x
